import logging

from sfdata_stream_parser.filters import generic

from sf_pipelines_test.common import stream_filters as stream_functions
from sf_pipelines_test.common.data import DataContainer, FileLocator, ProcessResult
from sf_pipelines_test.common.spec.__data_schema import DataSchema
from sf_pipelines_test.common.stream_pipeline import to_dataframe

from . import stream_filters

logger = logging.getLogger(__name__)


def task_cleanfile(src_file: FileLocator, schema: DataSchema) -> ProcessResult:
    """
    Clean input Annex A xlsx files according to schema and output clean data and errors
    :param src_file: The pointer to a file in a virtual filesystem
    :param schema: The data schema in a DataSchema class
    :return: A class containing a DataContainer and ErrorContainer
    """
    # Open & Parse file
    stream = stream_functions.tablib_parse(src_file)

    # Configure stream
    stream = stream_functions.add_table_name(stream, schema=schema)
    stream = stream_functions.inherit_property(stream, ["table_name", "table_spec"])
    stream = stream_filters.convert_column_header_to_match(stream, schema=schema)
    stream = stream_functions.match_config_to_cell(stream, schema=schema)

    # Clean stream
    stream = stream_functions.log_blanks(stream)
    stream = stream_functions.conform_cell_types(stream)

    # Create dataset
    stream = stream_functions.collect_cell_values_for_row(stream)
    dataset_holder, stream = stream_functions.collect_tables(stream)
    error_holder, stream = stream_functions.collect_errors(stream)

    # Consume stream so we know it's been processed
    generic.consume(stream)

    dataset = dataset_holder.value
    errors = error_holder.value

    logger.info(
        "Completed processing file %s with the following tables: %s",
        src_file.name,
        list(dataset.keys()),
    )

    dataset = DataContainer(
        {k: to_dataframe(v, schema.table[k]) for k, v in dataset.items()}
    )

    return ProcessResult(data=dataset, errors=errors)
